/*
Copyright (C) 2018-2019 de4dot@gmail.com

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

use super::super::super::test_utils::from_str_conv::*;
use super::super::super::MemorySize;
use super::super::enums::SymbolFlags;
use super::enums::OptionsProps;
use super::opt_value::OptionValue;
use super::options_parser::parse_option;
use super::sym_res_test_case::*;
#[cfg(not(feature = "std"))]
use alloc::string::String;
#[cfg(not(feature = "std"))]
use alloc::vec::Vec;
use core::iter::IntoIterator;
use core::u32;
#[cfg(not(feature = "std"))]
use hashbrown::HashMap;
#[cfg(feature = "std")]
use std::collections::HashMap;
use std::fs::File;
use std::io::prelude::*;
use std::io::{BufReader, Lines};
use std::path::Path;

pub(super) struct SymbolResolverTestParser {
	filename: String,
	lines: Lines<BufReader<File>>,
}

impl SymbolResolverTestParser {
	pub(super) fn new(filename: &Path) -> Self {
		let display_filename = filename.display().to_string();
		let file = File::open(filename).unwrap_or_else(|_| panic!("Couldn't open file {}", display_filename));
		let lines = BufReader::new(file).lines();
		Self { filename: display_filename, lines }
	}
}

impl IntoIterator for SymbolResolverTestParser {
	type Item = SymbolResolverTestCase;
	type IntoIter = IntoIter;

	fn into_iter(self) -> Self::IntoIter {
		// GENERATOR-BEGIN: OptionsDict
		// ⚠️This was generated by GENERATOR!🦹‍♂️
		let mut to_flags: HashMap<&'static str, u32> = HashMap::new();
		let _ = to_flags.insert("rel", SymbolFlags::RELATIVE);
		let _ = to_flags.insert("signed", SymbolFlags::SIGNED);
		// GENERATOR-END: OptionsDict

		IntoIter { filename: self.filename, lines: self.lines, line_number: 0, to_flags }
	}
}

pub(super) struct IntoIter {
	filename: String,
	lines: Lines<BufReader<File>>,
	line_number: u32,
	to_flags: HashMap<&'static str, u32>,
}

impl Iterator for IntoIter {
	type Item = SymbolResolverTestCase;

	fn next(&mut self) -> Option<Self::Item> {
		loop {
			match self.lines.next() {
				None => return None,
				Some(info) => {
					let result = match info {
						Ok(line) => {
							self.line_number += 1;
							if line.is_empty() || line.starts_with('#') {
								continue;
							}
							self.read_next_test_case(line)
						}
						Err(err) => Err(err.to_string()),
					};
					match result {
						Ok(tc) => return Some(tc),
						Err(err) => panic!("Error parsing symbol resolver test case file '{}', line {}: {}", self.filename, self.line_number, err),
					}
				}
			}
		}
	}
}

impl IntoIter {
	fn read_next_test_case(&self, line: String) -> Result<SymbolResolverTestCase, String> {
		let elems: Vec<_> = line.split(',').collect();
		const SYM_RES_INDEX: usize = 4;
		if elems.len() < SYM_RES_INDEX {
			return Err(format!("Invalid number of commas: {}", elems.len() - 1));
		}

		let bitness = to_u32(elems[0])?;
		let hex_bytes = String::from(elems[1].trim());
		let _ = to_vec_u8(&hex_bytes)?;
		let code = to_code(elems[2])?;

		let mut options: Vec<(OptionsProps, OptionValue)> = Vec::new();
		for value in elems[3].split_whitespace() {
			if value.is_empty() {
				continue;
			}
			options.push(parse_option(value)?);
		}

		let mut symbol_results: Vec<SymbolResultTestCase> = Vec::with_capacity(elems.len() - SYM_RES_INDEX);
		for tmp in &elems[SYM_RES_INDEX..] {
			let sym_parts: Vec<_> = tmp.split(';').collect();
			if sym_parts.len() != 5 {
				return Err(format!("Invalid number of semicolons: {}", sym_parts.len() - 1));
			}

			let address = to_u64(sym_parts[0])?;
			let symbol_address = to_u64(sym_parts[1])?;
			let address_size = to_u32(sym_parts[2])?;
			let symbol_parts: Vec<String> = sym_parts[3].split('|').map(String::from).collect();

			let mut memory_size: Option<MemorySize> = None;
			let mut flags = SymbolFlags::NONE;
			for value in sym_parts[4].split_whitespace() {
				if value.is_empty() {
					continue;
				}
				if let Some(f) = self.to_flags.get(value) {
					flags |= *f;
				} else {
					memory_size = Some(to_memory_size(value)?);
				}
			}
			symbol_results.push(SymbolResultTestCase { address, symbol_address, address_size, flags, memory_size, symbol_parts })
		}

		Ok(SymbolResolverTestCase { bitness, hex_bytes, code, options, symbol_results })
	}
}
